-- Game report
--require("datastream")
require("glon")

include("cl_awards.lua")

CLSCORE = {}
CLSCORE.Events = {}
CLSCORE.Scores = {}
CLSCORE.TraitorIDs = {}
CLSCORE.Players = {}
CLSCORE.StartTime = 0
CLSCORE.Panel = nil

local skull_icon = Material("HUD/killicons/default")
local join_icon = Material("gui/silkicons/add")
local smile_icon = Material("gui/silkicons/emoticon_smile")
local magnifier_icon = Material("gui/silkicons/magnifier")
--local bomb_icon = Material("gui/silkicons/bomb")
local wrong_icon = Material("gui/silkicons/check_off")
local right_icon = Material("gui/silkicons/check_on")
local shield_icon = Material("gui/silkicons/shield")
local star_icon = Material("gui/silkicons/star")
local app_icon = Material("gui/silkicons/application")

surface.CreateFont("Trebuchet24", 72, 1000, true, false, "WinHuge")

function CLSCORE:TextForEvent(e)
   if e.id == EVENT_KILL then
      if e.att.uid == e.vic.uid then
         return string.format("%s couldn't take the strain and killed himself", e.vic.ni)
      end

      if e.att.uid == -1 then
         return string.format("Something killed %s", e.vic.ni)
      end

      local dmg = e.dmg
      if util.HasDamageBit(dmg.t, DMG_FALL) then
         if e.att.ni != "" then
            return string.format("%s fell to his death after a push from %s", e.vic.ni, e.att.ni)
         else
            return string.format("%s fell to his death", e.vic.ni)
         end
      elseif util.HasDamageBit(dmg.t, DMG_BULLET) then
         local gun = EnumToWep(dmg.g)
         if gun != nil then
            return string.format("%s killed %s using a %s", e.att.ni, e.vic.ni, gun)
         else
            return string.format("%s killed %s", e.att.ni, e.vic.ni)
         end
      elseif util.HasDamageBit(dmg.t, DMG_DROWN) then
         return string.format("%s drowned", e.vic.ni)
      elseif util.HasDamageBit(dmg.t, DMG_BLAST) then
         return string.format("%s was exploded by %s", e.vic.ni, e.att.ni)
      elseif util.HasDamageBit(dmg.t, DMG_BURN) then
         return string.format("%s was fried by %s", e.vic.ni, e.att.ni)
      elseif util.HasDamageBit(dmg.t, DMG_DIRECT) then
         return string.format("%s was slowly burned by %s", e.vic.ni, e.att.ni)
      elseif util.HasDamageBit(dmg.t, DMG_CLUB) then
         return string.format("%s was beaten up by %s", e.vic.ni, e.att.ni)
      elseif util.HasDamageBit(dmg.t, DMG_SLASH) then
         return string.format("%s was stabbed to death by %s", e.vic.ni, e.att.ni)
      else
         return string.format("%s killed %s", e.att.ni, e.vic.ni)
      end
   elseif e.id == EVENT_BODYFOUND then
      return string.format("%s found the corpse of %s", e.ni, e.b)
   elseif e.id == EVENT_SPAWN then
      return string.format("%s joined the fun", e.ni)
   elseif e.id == EVENT_GAME then
      if e.state == ROUND_ACTIVE then
         return "The round began"
      elseif e.state == ROUND_PREP then
         return "The preparation phase started"
      elseif e.state == ROUND_POST then
         return "The post-round fun got going"
      end
   elseif e.id == EVENT_FINISH then
      if e.win == WIN_TRAITOR then
         return "The traitors won the round!"
      elseif e.win == WIN_INNOCENT then
         return "The pure and innocent terrorists won the round!"
      elseif e.win == WIN_TIMELIMIT then
         return "The traitors lost when the timelimit was reached!"
      else
         return "The round ended"
      end
   end

   return nil
end

function CLSCORE:IconForEvent(e)
   if e.id == EVENT_KILL then
      if e.att.uid == e.vic.uid or e.att.uid == -1 then
         return smile_icon, "Suicide"
      end

      if e.att.tr == e.vic.tr then
         return wrong_icon, "Teamkill"
      elseif e.att.tr then
         return right_icon, "Traitor killed innocent"
      else
         return shield_icon, "Innocent killed traitor"
      end
   elseif e.id == EVENT_SPAWN then
      return join_icon, "Player joined"
   elseif e.id == EVENT_GAME then
      return app_icon, "Game status changed"
   elseif e.id == EVENT_FINISH then
      if e.win == WIN_TRAITOR then
         return star_icon, "Traitors won"
      elseif e.win == WIN_INNOCENT then
         return star_icon, "Innocents won"
      elseif e.win == WIN_TIMELIMIT then
         return star_icon, "Innocents won by default"
      else
         return star_icon, "Innocents won"
      end
   elseif e.id == EVENT_BODYFOUND then
      return magnifier_icon, "Body discovered"
   end

   return nil
end

function CLSCORE:FillDList(dlst)
   local starttime = self.StartTime
   for k, e in pairs(self.Events) do

      local etxt = self:TextForEvent(e)
      local eicon, ttip = self:IconForEvent(e)
      local etime = ""
      local t = e.t - starttime
      if t >= 0 then
         etime = string.FormattedTime(t, "%02i:%02i")
      end
      
      if etxt then
         if eicon then
            local mat = eicon
            eicon = vgui.Create("DImage")
            eicon:SetMaterial(mat)
            eicon:SetTooltip(ttip)
            eicon:SetKeepAspect(true)
            eicon:SetSize(16,16)
         end
         dlst:AddLine(etime, eicon, "  " .. etxt)
      end
   end
end

function CLSCORE:BuildEventLogPanel(dpanel)
   local margin = 10

   local w, h = dpanel:GetSize()

   local dlist = vgui.Create("DListView", dpanel)
   dlist:SetPos(0, 2)
   dlist:SetSize(w, h - 25)
   dlist:SetSortable(true)
   dlist:SetMultiSelect(false)

   local timecol = dlist:AddColumn("Time")
   local iconcol = dlist:AddColumn("")
   local eventcol = dlist:AddColumn("Event")

   iconcol:SetFixedWidth(16)
   timecol:SetFixedWidth(40)

   -- If sortable is off, no background is drawn for the headers which looks
   -- terrible. So enable it, but disable the actual use of sorting.
   iconcol.Header:SetDisabled(true)
   timecol.Header:SetDisabled(true)
   eventcol.Header:SetDisabled(true)

   self:FillDList(dlist)

--   dtab1:SizeToContents()
end

function CLSCORE:BuildScorePanel(dpanel)
   local margin = 10

   local w, h = dpanel:GetSize()

   local dlist = vgui.Create("DListView", dpanel)
   dlist:SetPos(0, 0)
   dlist:SetSize(w, h)
   dlist:SetSortable(true)
   dlist:SetMultiSelect(false)

   local cols = {}
   local colnames = {"", "Player", "Role", "Innocent kills", "Traitor kills", "Points", "Team bonus", "Total points"}
   for k, name in pairs(colnames) do
      local c = dlist:AddColumn(name)

      -- size for icon
      if name == "" then
         c:SetFixedWidth(22)
      end
         
   end

   local scores = self.Scores
   local nicks = self.Players
   local bonus = ScoreTeamBonus(scores)

   for id, s in pairs(scores) do
      local was_traitor = s.was_traitor
      local role = was_traitor and "Traitor" or ""
--      local surv = s.deaths > 0 and "DEAD" or ""
      local surv = ""
      if s.deaths > 0 then
         surv = vgui.Create("DImage", dlist)
         surv:SetMaterial(skull_icon)
         surv:SetTooltip("Died")
         surv:SetKeepAspect(true)
         surv:SetSize(16,16)
      end

      local points_own = KillsToPoints(s, was_traitor)
      local points_team = (was_traitor and bonus.traitors or bonus.innos)
      local points_total = points_own + points_team

      local l = dlist:AddLine(surv, nicks[id], role, s.innos, s.traitors, points_own, points_team, points_total)

      for k, col in pairs(l.Columns) do
         col:SetContentAlignment(5)
      end
   end

   dlist:SortByColumn(6)

end

function CLSCORE:AddAward(y, pw, award, dpanel)
   local nick = award.nick
   local text = award.text
   local title = string.upper(award.title)

   local titlelbl = vgui.Create("Label", dpanel)
   titlelbl:SetText(title)
   titlelbl:SetFont("TabLarge")
   titlelbl:SizeToContents()
   local tiw, tih = titlelbl:GetSize()

   local nicklbl = vgui.Create("Label", dpanel)
   nicklbl:SetText(nick)
   nicklbl:SetFont("DefaultBold")
   nicklbl:SizeToContents()
   local nw, nh = nicklbl:GetSize()
   
   local txtlbl = vgui.Create("Label", dpanel)
   txtlbl:SetText(text)
   txtlbl:SetFont("Default")
   txtlbl:SizeToContents()
   local tw, th = txtlbl:GetSize()

   titlelbl:SetPos((pw - tiw) / 2, y)
   y = y + tih + 2

   local fw = nw + tw + 5
   local fx = ((pw - fw) / 2)
   nicklbl:SetPos(fx, y)
   txtlbl:SetPos(fx + nw + 5, y)

   y = y + nh

   return y
end

-- double check that we have no nils
local function ValidAward(a)
   return a and a.nick and a.text and a.title and a.priority
end

function CLSCORE:BuildHilitePanel(dpanel)
   local w, h = dpanel:GetSize()

   local wintext = "THE INNOCENT WIN"
   local endtime = self.StartTime
   for k, e in pairs(self.Events) do
      if e.id == EVENT_FINISH then
         endtime = e.t

         if e.win == WIN_TRAITOR then
            wintext = "THE TRAITORS WIN"
         end
      end
   end
   local roundtime = endtime - self.StartTime

   local numply = table.Count(self.Players)
   local numtr = table.Count(self.TraitorIDs)


   local bg = vgui.Create("DColouredBox", dpanel)
   bg:SetColor(Color(50, 50, 50, 255))
   bg:SetSize(w,h)
   bg:SetPos(0,0)

   local winlbl = vgui.Create("Label", dpanel)
   winlbl:SetFont("WinHuge")
   winlbl:SetText(wintext)
   winlbl:SizeToContents()
   local xwin = (w - winlbl:GetWide())/2
   local ywin = 35
   winlbl:SetPos(xwin, ywin)

   local ysubwin = ywin + winlbl:GetTall()
   local partlbl = vgui.Create("DLabel", dpanel)

   local plytxt = numply .. " players took part, " .. numtr .. " were traitors."
   if numtr < 2 then plytxt = numply .. " players took part, one of them a traitor." end

   partlbl:SetText(plytxt)
   partlbl:SizeToContents()
   partlbl:SetPos(xwin, ysubwin)

   local timelbl = vgui.Create("DLabel", dpanel)
   timelbl:SetText("The round lasted " .. string.FormattedTime(roundtime, "%02i:%02i") .. ".")
   timelbl:SizeToContents()
   timelbl:SetPos(xwin + winlbl:GetWide() - timelbl:GetWide(), ysubwin)

   -- Awards
   local wa = math.Round(w * 0.9)
   local ha = h - ysubwin - 40
   local xa = (w - wa) / 2
   local ya = h - ha

   local awardp = vgui.Create("DPanel", dpanel)
   awardp:SetSize(wa, ha)
   awardp:SetPos(xa, ya)
   awardp:SetPaintBackground(false)

   -- Before we pick awards, seed the rng in a way that is the same on all
   -- clients. We can do this using the round start time. To make it a bit more
   -- random, involve the round's duration too.
   math.randomseed(self.StartTime + endtime)

   -- Attempt to generate every award, then sort the succeeded ones based on
   -- priority/interestingness
   local award_choices = {}
   for k, afn in pairs(AWARDS) do
      local a = afn(self.Events, self.Scores, self.Players, self.TraitorIDs)
      if ValidAward(a) then
         table.insert(award_choices, a)
      end
   end

   local num_choices = table.Count(award_choices)
   local max_awards = 5

   -- sort descending by priority
   table.SortByMember(award_choices, "priority")

   -- put the N most interesting awards in the menu
   for i=1,max_awards do
      local a = award_choices[i]
      if a then
         self:AddAward((i - 1) * 42, wa, a, awardp)
      end
   end
end

function CLSCORE:ShowPanel()
   local margin = 15

   local dpanel = vgui.Create("DFrame")
   local w, h = 700, 500
   dpanel:SetSize(700, 500)
   dpanel:Center()
   dpanel:SetTitle("Round report")
   dpanel:SetVisible(true)
   dpanel:ShowCloseButton(true)
   dpanel:SetMouseInputEnabled(true)

   -- keep it around so we can reopen easily
   dpanel:SetDeleteOnClose(false)
   self.Panel = dpanel

   local dbut = vgui.Create("DButton", dpanel)
   local bw, bh = 50, 25
   dbut:SetSize(bw, bh)
   dbut:SetPos(w - bw - margin, h - bh - margin/2)
   dbut:SetText("Close")
   dbut.DoClick = function() dpanel:Close() end

   local dtabsheet = vgui.Create("DPropertySheet", dpanel)
   dtabsheet:SetPos(margin, margin + 15)
   dtabsheet:SetSize(w - margin*2, h - margin*3 - bh)
   local padding = dtabsheet:GetPadding()


   -- Highlight tab
   local dtabhilite = vgui.Create("DPanel", dtabsheet)
   dtabhilite:SetPaintBackground(false)
   dtabhilite:StretchToParent(padding,padding,padding,padding)
   self:BuildHilitePanel(dtabhilite)

   dtabsheet:AddSheet("Highlights", dtabhilite, "gui/silkicons/star", false, false, "Round highlights")

   -- Event log tab
   local dtabevents = vgui.Create("DPanel", dtabsheet)
--   dtab1:SetSize(650, 450)
   dtabevents:StretchToParent(padding, padding, padding, padding)
   self:BuildEventLogPanel(dtabevents)

   dtabsheet:AddSheet("Events", dtabevents, "gui/silkicons/application_view_detail", false, false, "Log of the events that happened this round")

   -- Score tab
   local dtabscores = vgui.Create("DPanel", dtabsheet)
   dtabscores:SetPaintBackground(false)
   dtabscores:StretchToParent(padding, padding, padding, padding)
   self:BuildScorePanel(dtabscores)

   dtabsheet:AddSheet("Scores", dtabscores, "gui/silkicons/user", false, false, "Points scored by each player in this round alone")

   dpanel:MakePopup()

   -- makepopup grabs keyboard, whereas we only need mouse
   dpanel:SetKeyboardInputEnabled(false)
end

function CLSCORE:ClearPanel()

   if self.Panel then
      -- move the mouse off any tooltips and then remove the panel next tick

      -- we need this hack as opposed to just calling Remove because gmod does
      -- not offer a means of killing the tooltip, and doesn't clean it up
      -- properly on Remove
      gui.SetMousePos( ScrW()/2, ScrH()/2 )
      timer.Simple(0, self.Panel.Remove, self.Panel)
   end
end

function CLSCORE:Reset()
   self.Events = {}
   self.StoredEvents = nil
   self.TraitorIDs = {}
   self.Scores = {}
   self.Players = {}
   self.RoundStarted = 0

   self:ClearPanel()
end

function CLSCORE:Init(events)
   -- Get start time and traitors
   local starttime = nil
   local traitors = nil
   for k, e in pairs(events) do
      if e.id == EVENT_GAME and e.state == ROUND_ACTIVE then
         starttime = e.t
      elseif e.id == EVENT_SELECTED then
         traitors = e.traitor_ids
      end

      if starttime and traitors then
         break
      end
   end
   
   -- Get scores and players
   local scores = {}
   local nicks = {}
   for k, e in pairs(events) do
      if e.id == EVENT_SPAWN then
         scores[e.uid] = ScoreInit()
         nicks[e.uid] = e.ni
      end
   end

   scores = ScoreEventLog(events, scores)

   self.Players = nicks
   self.Scores = scores
   self.TraitorIDs = traitors
   self.StartTime = starttime
   self.Events = events
end

function CLSCORE:ReportEvents(events)
   self:Reset()

--   if events != self.StoredEvents then
--      self.StoredEvents = table.Copy(events)
--   end

   self:Init(events)
   self:ShowPanel()
end

function CLSCORE:Reopen()
   if self.Panel and self.Panel:IsValid() and not self.Panel:IsVisible() then
      self.Panel:SetVisible(true)
   end
end

local buff = ""
--evdebug = nil
local function ReceiveReportStream(um)
--   Msg("Receiving message... \n")
   local cont = um:ReadBool()

   buff = buff .. um:ReadString()

   if cont then 
      return
   else
--      Msg("MESSAGE COMPLETE, length: ".. #buff .. "\n") 

      -- do stuff with buffer contents
      local status, results = pcall(glon.decode, buff)

      if status then
--         Msg("Decoded: \n")
--         PrintTable(results)
         CLSCORE:ReportEvents(results)
      else
         Msg("Round report event encoding returned an error: " .. results .. "\n")
      end

      -- flush
      buff = ""      
   end


end
usermessage.Hook("report_stream", ReceiveReportStream)

